from __future__ import annotations

from types import MappingProxyType

from ._base import Command
from ._filter import Filter
from ._history import History
from ._plot import Plot
from ._suggest import Suggest
from ._sync import Sync
from ._top_files import TopFiles
from ._version import Version


commands: MappingProxyType[str, type[Command]]
commands = MappingProxyType({
    'filter': Filter,
    'history': History,
    'plot': Plot,
    'suggest': Suggest,
    'sync': Sync,
    'top-files': TopFiles,
    'version': Version,
})

__all__ = [
    'commands',
    'Command',
]



project = 'mypy-baseline'
copyright = '2023, @orsinium'
author = '@orsinium'
templates_path = ['_templates']
html_theme = 'alabaster'
autodoc_typehints_format = 'short'
autodoc_preserve_defaults = True
autodoc_member_order = 'bysource'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'myst_parser',
]



from __future__ import annotations

import sys
from argparse import ArgumentParser
from typing import NoReturn, TextIO

from .commands import Command, commands


def main(argv: list[str], stdin: TextIO, stdout: TextIO) -> int:
    parser = ArgumentParser('mypy-baseline')
    subparsers = parser.add_subparsers()
    parser.set_defaults(cmd=None)

    cmd_class: type[Command]
    for name, cmd_class in sorted(commands.items()):
        subparser = subparsers.add_parser(name=name, help=cmd_class.__doc__)
        subparser.set_defaults(cmd=cmd_class)
        cmd_class.init_parser(subparser)
    args = parser.parse_args(argv)

    cmd_class = args.cmd
    if cmd_class is None:
        parser.print_help()
        return 1
    cmd = cmd_class(args=args, stdin=stdin, stdout=stdout)
    return cmd.run()


def entrypoint() -> NoReturn:
    sys.exit(main(argv=sys.argv[1:], stdin=sys.stdin, stdout=sys.stdout))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


RED = '\033[31m'
GREEN = '\033[32m'
BLUE = '\033[94m'
MAGENTA = '\033[35m'
END = '\033[0m'

NEW_ERRORS = """
    ┌────────────────────────────────────────────┄┄
    │ {red}Your changes introduced new violations.{end}
    │ Please, resolve the violations above before moving forward.
    │ Mypy is your friend.
    └────────────────────────────────────────────┄┄
"""
FIXED_ERRORS = """
    ┌────────────────────────────────────────────┄┄
    │ {green}Your changes resolved existing violations.{end}
    │ Great work! Please, run `mypy | mypy-baseline sync`
    │ to remove resolved violations from the baseline file.
    └────────────────────────────────────────────┄┄
"""


@dataclass
class Colors:
    disabled: bool

    def green(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{GREEN}{text}{END}'

    def red(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{RED}{text}{END}'

    def blue(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{BLUE}{text}{END}'

    def magenta(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{MAGENTA}{text}{END}'

    def get_exit_message(self, fixed: int, new: int) -> str:
        if new:
            msg = NEW_ERRORS
        elif fixed:
            msg = FIXED_ERRORS
        else:
            return ''
        if self.disabled:
            red = ''
            green = ''
            end = ''
        else:
            red = RED
            green = GREEN
            end = END
        return msg.format(
            red=red,
            green=green,
            end=end,
        )



from __future__ import annotations

from pathlib import Path

from .helpers import LINE1, LINE2, run


def test_top_files(tmp_path: Path):
    blpath = tmp_path / 'bline.txt'
    blpath.write_text(f'{LINE1}{LINE2}')
    cmd = ['top-files', '--baseline-path', str(blpath), '--no-color']
    actual = run(cmd)
    assert 'views.py' in actual
    assert 'settings.py' in actual



from __future__ import annotations

import os
import random
import subprocess
from argparse import ArgumentParser
from functools import cached_property
from pathlib import Path
from textwrap import dedent
from typing import Iterable

from .._config import Config
from .._error import Error
from ._base import Command


TARGET_BRANCH_ENV_VARS = (
    # GitLab CI
    'CI_MERGE_REQUEST_TARGET_BRANCH_SHA',
    'CI_MERGE_REQUEST_TARGET_BRANCH_NAME',
    # GitHub Actions
    'GITHUB_BASE_REF',
)
PR_ID_ENV_VARS = (
    # GitLab CI
    'CI_MERGE_REQUEST_ID',
    # GitHub Actions
    'GITHUB_REF_NAME',
)


class Suggest(Command):
    """Suggest to fix a violation from the baseline.
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        Command.init_parser(parser)
        parser.add_argument(
            '--seed',
            help='seed to use when randomly picking a suggestion',
        )
        parser.add_argument(
            '--min-fixed', default=1,
            help='required number of fixes for the MR',
        )
        parser.add_argument(
            '--exit-zero', action='store_true',
            help='always return zero exit code',
        )
        parser.add_argument(
            '--comment', action='store_true',
            help='add comment to the GitLab MR discussion',
        )

    def run(self) -> int:
        if self.fixed_count >= self.args.min_fixed:
            return 0
        config = Config(preserve_position=True)
        result = self.suggested.get_clean_line(config)
        self.print(result)
        if self.args.comment:
            self._post_to_gitlab(result)
        if self.args.exit_zero:
            return 0
        return 1

    @cached_property
    def target(self) -> str:
        """Get the target branch/commit reference for this PR.
        """
        if self.config.default_branch:
            return self.config.default_branch

        # detect default branch from env vars
        for env_var in TARGET_BRANCH_ENV_VARS:
            target = os.environ.get(env_var)
            if target:
                return target

        # detect default branch for the `origin` remote
        # (may fail if there is no remote)
        try:
            lines = self._get_stdout('symbolic-ref', 'refs/remotes/origin/HEAD')
        except subprocess.CalledProcessError:
            pass
        else:
            return lines[-1].split('/')[-1]

        # try well-known branch names
        for branch_name in ('main', 'master', 'develop'):
            res = subprocess.run(
                ['git', 'rev-parse', '--verify', branch_name],
                stdout=subprocess.DEVNULL,
            )
            if res.returncode == 0:
                return branch_name
        raise LookupError('cannot detect default branch name')

    @cached_property
    def fixed_count(self) -> int:
        """Get number of baseline violations fixed in this PR.
        """
        lines = self._get_stdout(
            'diff', 'HEAD', self.target, '--',
            str(self.config.baseline_path),
        )
        count = 0
        for line in lines:
            if line.startswith('-') and not line.startswith('---'):
                count += 1
        return count

    @cached_property
    def changed_files(self) -> tuple[Path, ...]:
        """Get all lines changed in this PR.
        """
        lines = self._get_stdout('diff', '--name-only', self.target)
        return tuple(Path(line).absolute() for line in lines)

    def _get_stdout(self, *args: str) -> list[str]:
        """Run the command in the shell and get stdout lines.
        """
        result = subprocess.run(['git', *args], stdout=subprocess.PIPE)
        result.check_returncode()
        stdout = result.stdout.decode().strip()
        return stdout.splitlines()

    @property
    def suggested(self) -> Error:
        """Pick a violation that shoud be suggested to fix.
        """
        errors = [e for e in self.baseline if e.severity == 'error']
        if not errors:
            errors = self.baseline
        # pick an error in one of the changed files
        for error in errors:
            if error.path.absolute() in self.changed_files:
                return error
        # pick a random error
        random.seed(self.seed)
        return random.choice(errors)

    @cached_property
    def baseline(self) -> list[Error]:
        """Parse errors from the baseline.
        """
        baseline: list[Error] = []
        for line in self.baseline_lines:
            err = Error.new(line)
            if err is not None:
                baseline.append(err)
        if not baseline:
            raise RuntimeError('baseline is empty')
        return baseline

    @property
    def baseline_lines(self) -> Iterable[str]:
        """Read baseline and return lines of it, one per violation.
        """
        # if not self.stdin.isatty():
        #     return self.stdin
        text = self.config.baseline_path.read_text(encoding='utf8')
        return text.splitlines()

    @cached_property
    def seed(self) -> str:
        """Get the seed to init randomizer.

        If `--seed` is not specified, try to pick value
        so that it's always the same for the same MR
        but different for different MRs.
        """
        if self.args.seed:
            return self.args.seed
        for env_var in PR_ID_ENV_VARS:
            pr_id = os.environ.get(env_var)
            if pr_id:
                return pr_id
        return ''

    def _post_to_gitlab(self, err_msg: str) -> None:
        """Add a comment to the GitLab MR for which the job runs.
        """
        import requests

        # get API token
        is_gitlab = os.environ.get('GITLAB_CI')
        if not is_gitlab:
            raise LookupError('Not running on GitLab CI')
        token = os.environ.get('REVIEWDOG_GITLAB_API_TOKEN')
        if not token:
            token = os.environ.get('GITLAB_API_TOKEN')
        if not token:
            raise LookupError('GITLAB_API_TOKEN env var is not set')

        already_commented = self._check_gitlab_has_comment(token)
        if already_commented:
            return

        # form the message
        s = 's' if self.args.min_fixed > 1 else ''
        path = self.config.baseline_path
        count = self.args.min_fixed
        msg = f"""
            ## mypy-baseline suggest

            Please, resolve at least {count} error{s} from `{path}`. Suggested:

            ```
            {err_msg}
            ```
        """

        # send the request
        resp = requests.post(
            url=self._gitlab_comment_url,
            json={'body': dedent(msg)},
            headers={'PRIVATE-TOKEN': token},
        )
        resp.raise_for_status()

    def _check_gitlab_has_comment(self, token: str) -> bool:
        """Check if mypy-baseline already left a comment in the MR.
        """
        import requests

        resp = requests.get(self._gitlab_comment_url)
        if not resp.ok:
            return False
        for comment in resp.json():
            body: str = comment['body'].strip()
            if body.lstrip().startswith('## mypy-baseline suggest'):
                return True
        return False

    @cached_property
    def _gitlab_comment_url(self) -> str:
        """GitLab API endpoint to work with MR comments.
        """
        base_url = os.environ['CI_API_V4_URL']
        project_id = os.environ['CI_MERGE_REQUEST_PROJECT_ID']
        mr_id = os.environ['CI_MERGE_REQUEST_IID']
        return f'{base_url}/projects/{project_id}/merge_requests/{mr_id}/notes'



from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path

from .._git import get_commits
from ._base import Command


class Plot(Command):
    """Draw the graph of how the baseline changed over time.

    Requires plotnine to be installed.
    """
    @classmethod
    def init_parser(cls, parser: ArgumentParser) -> None:
        super().init_parser(parser)
        parser.add_argument(
            '--output', type=Path, default=Path('mypy-baseline.png'),
            help='the path to save the file to',
        )

    def run(self) -> int:
        import pandas
        import plotnine as gg
        commits = list(get_commits(self.config.baseline_path))

        prev_count: int | None = None
        for commit in commits:
            commit.fix_lines_count(prev_count)
            prev_count = commit.lines_count

        commits = commits[1:]  # drop the oldest commit
        df = pandas.DataFrame(c.as_dict() for c in commits)
        df['created_at'] = pandas.to_datetime(df.created_at, utc=True)
        graph = (
            gg.ggplot(df, gg.aes(x='created_at', y='lines_count'))
            + gg.geom_line()
            + gg.geom_point(gg.aes(color='deletions > insertions'))
            + gg.ylim(0, max(c.lines_count for c in commits) + 5)
            + gg.theme(axis_text_x=gg.element_text(rotation=45, hjust=1))
            + gg.xlab('commit time')
            + gg.ylab('unresolved issues')
        )
        graph.save(self.args.output)
        return 0



from __future__ import annotations

import re
import subprocess
from dataclasses import asdict, dataclass
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import Iterator


REX_COMMIT = re.compile(
    (
        r'(?P<hash>[0-9a-f]{40}) '
        r'(?P<created_at>[0-9T:+-]+)\s+'
        r'(?P<author_email>[^\s]+)\s+'
        r'.+ \| \d+ [+-]+\s+'
        r'\d+ files? changed'
        r'(?:, (?P<insertions>\d+) insertions?\(\+\))?'
        r'(?:, (?P<deletions>\d+) deletions?\(\-\))?'
    ),
    re.MULTILINE
)


@dataclass
class Commit:
    path: Path
    hash: str
    author_email: str
    created_at: datetime
    insertions: int
    deletions: int

    @cached_property
    def lines_count(self) -> int:
        path = self.path
        if path.is_absolute():
            path = path.relative_to(Path().absolute())
        cmd = ['git', 'show', f'{self.hash}:{path}']
        result = subprocess.run(cmd, stdout=subprocess.PIPE)
        result.check_returncode()
        lines = result.stdout.decode().strip().splitlines()
        return len(lines)

    def fix_lines_count(self, prev_count: int | None) -> None:
        """
        We want to show not the actual lines count that the commit resulted in
        but how its changed affected the previous count.
        These numbers can be different if there are 2 commits with the same
        parent commit that changed the baseline.
        """
        if prev_count is not None:
            self.lines_count = prev_count + self.insertions - self.deletions

    def as_dict(self) -> dict[str, object]:
        result = asdict(self)
        result['lines_count'] = self.lines_count
        return result


def get_commits(path: Path) -> Iterator[Commit]:
    cmd = ['git', 'log', '--format=%H %cI %ae', '--reverse', '--stat', '--', str(path)]
    result = subprocess.run(cmd, stdout=subprocess.PIPE)
    result.check_returncode()
    stdout = result.stdout.decode()
    for match in REX_COMMIT.finditer(stdout):
        info = match.groupdict()
        yield Commit(
            path=path,
            hash=info['hash'],
            author_email=info['author_email'],
            created_at=datetime.fromisoformat(info['created_at']),
            insertions=int(info['insertions'] or '0'),
            deletions=int(info['deletions'] or '0'),
        )



from __future__ import annotations

from .._error import Error
from ._base import Command


class Sync(Command):
    """Generate baseline with all existing errors.
    """

    def run(self) -> int:
        try:
            baseline_text = self.config.baseline_path.read_text(encoding='utf8')
        except FileNotFoundError:
            baseline_text = ''
        old_baseline = baseline_text.splitlines()

        new_baseline: list[str] = []
        for line in self.stdin:
            error = Error.new(line)
            if error is None:
                self.print(line, end='')
                continue
            if self.config.is_ignored(error.message):
                continue
            clean_line = error.get_clean_line(self.config)
            new_baseline.append(clean_line)

        synced = False
        if old_baseline:
            synced = self._stable_sync(old_baseline, new_baseline)
        if not synced:
            self._write_baseline(new_baseline)
        return 0

    def _stable_sync(self, old_bline: list[str], new_bline: list[str]) -> bool:
        """Try to cleanly update the old baseline instead of rewriting it.

        We do that to avoid lines jumping around that happens because of not ordered
        output from mypy.

        Currently, we can do a stable sync only when there are no new lines added.
        It's hard to insert new lines in the correct positions, and adding them
        at the end of the file will cause merge conflicts.
        Sorting lines alphabetically would solve the issue, but I want to keep
        backward compatibility.
        """
        old_set = set(old_bline)
        new_set = set(new_bline)
        removed = old_set - new_set
        added = new_set - old_set
        if added:
            return False
        result = [line for line in old_bline if line not in removed]
        self._write_baseline(result)
        return True

    def _write_baseline(self, baseline: list[str]) -> None:
        """Serialize baseline and write it into the file.
        """
        serialized = '\n'.join(baseline) + '\n'
        self.config.baseline_path.write_text(serialized, encoding='utf8')



from __future__ import annotations

import dataclasses
import os
import re
from argparse import ArgumentParser
from functools import cached_property
from pathlib import Path
from typing import Any


try:
    import toml
except ImportError:
    toml = None  # type: ignore[assignment]
try:
    import tomli
except ImportError:
    # https://peps.python.org/pep-0680/
    try:
        import tomllib as tomli  # type: ignore
    except ImportError:
        tomli = None   # type: ignore[assignment]


@dataclasses.dataclass
class Config:
    baseline_path: Path = Path('mypy-baseline.txt')
    depth: int = 40
    allow_unsynced: bool = False
    preserve_position: bool = False
    hide_stats: bool = False
    no_colors: bool = bool(os.environ.get('NO_COLOR'))
    ignore: list[str] = dataclasses.field(default_factory=list)
    default_branch: str = ''

    @classmethod
    def from_args(cls, args: dict[str, Any]) -> Config:
        config = cls()
        config = config.read_file(args['config'])
        config = config.read_args(args)
        if isinstance(config.baseline_path, str):
            config.baseline_path = Path(config.baseline_path)
        return config

    @classmethod
    def init_parser(self, parser: ArgumentParser) -> None:
        add = parser.add_argument
        add(
            '--config', type=Path, default=Path('pyproject.toml'),
            help='path to the configuration file.'
        )
        add(
            '--baseline-path', type=Path,
            help='path to the file where to store baseline.',
        )
        add(
            '--depth', type=int,
            help='cut paths longer than that many directories deep.'
        )

        add(
            '--allow-unsynced', action='store_true',
            help='do not fail for resolved violations.'
        )
        add(
            '--preserve-position', action='store_true',
            help='do not remove line number from the baseline.',
        )
        add(
            '--hide-stats', action='store_true',
            help='do not show stats at the end.',
        )
        add(
            '--no-colors', action='store_true',
            help='disable colored output. Has no effect on the output of mypy.',
        )
        add(
            '--ignore', nargs='*',
            help='regexes for messages to ignore.',
        )
        add(
            '--default-branch',
            help='default git branch name.',
        )

    def read_file(self, path: Path) -> Config:
        if not path.exists():
            return self

        # parse the config file
        if tomli is not None:
            with path.open('rb') as stream:
                data = tomli.load(stream)
        elif toml is not None:
            with path.open('rb', encoding='utf8') as stream:
                data = dict(toml.load(stream))
        else:
            return self

        # extract the right section from the config
        try:
            data = data['tool']['mypy-baseline']
        except KeyError:
            try:
                data = data['tool']['mypy_baseline']
            except KeyError:
                return self
        return dataclasses.replace(self, **data)

    def read_args(self, args: dict[str, Any]) -> Config:
        config = dataclasses.replace(self)
        for field in dataclasses.fields(self):
            value = args[field.name]
            if value is not None:
                setattr(config, field.name, value)
        return config

    @cached_property
    def _ignore_regexes(self) -> list[re.Pattern[str]]:
        result = []
        assert isinstance(self.ignore, list)
        for pattern in self.ignore:
            try:
                result.append(re.compile(pattern))
            except re.error:
                raise ValueError(f'Invalid pattern: {pattern}')
        return result

    def is_ignored(self, msg: str) -> bool:
        """Check if the message matches any ignore pattern from the config.
        """
        return any(rex.fullmatch(msg) for rex in self._ignore_regexes)



from mypy_baseline._config import Config
from mypy_baseline._error import Error


LINE1 = 'my_project/api/views.py:69: \x1b[1m\x1b[31merror:\x1b[m\x0f Incompatible types in assignment (expression has type \x1b[m\x0f\x1b[1m"List[Type[RawRenderer]]"\x1b[m\x0f, base class \x1b[m\x0f\x1b[1m"BaseAPI"\x1b[m\x0f defined the type as \x1b[m\x0f\x1b[1m"Tuple[Any, Type[RawRenderer]]"\x1b[m\x0f)  \x1b[m\x0f\x1b[33m[assignment]\x1b[m\x0f\r\n'  # noqa
LINE1EXP = 'my_project/api/views.py:0: error: Incompatible types in assignment (expression has type "List[Type[RawRenderer]]", base class "BaseAPI" defined the type as "Tuple[Any, Type[RawRenderer]]")  [assignment]'  # noqa


def test_line1_parse():
    e = Error.new(LINE1)
    assert e is not None
    assert e.path.parts == ('my_project', 'api', 'views.py')
    assert e.line_number == 69
    assert e.severity == 'error'
    assert e.message.startswith('Incompatible types in')
    assert e.category == 'assignment'
    assert e.get_clean_line(Config()) == LINE1EXP


LINE2 = 'my_project/api/views.py:0: note: This violates the Liskov substitution principle\r\n'  # noqa
LINE2EXP = 'my_project/api/views.py:0: note: This violates the Liskov substitution principle'  # noqa


def test_line2_parse():
    e = Error.new(LINE2)
    assert e is not None
    assert e.path.parts == ('my_project', 'api', 'views.py')
    assert e.line_number == 0
    assert e.severity == 'note'
    assert e.message == 'This violates the Liskov substitution principle'
    assert e.category == 'note'
    assert e.get_clean_line(Config()) == LINE2EXP
